@@ -78,6 +78,8 @@ class RoborockStateCode(RoborockEnum):
     mapping = 29
     egg_attack = 30
     patrol = 32
+    attaching_the_mop = 33  # on g20s ultra
+    detaching_the_mop = 34  # on g20s ultra
     charging_complete = 100
     device_offline = 101
     locked = 103
@@ -259,7 +261,9 @@ class RoborockFanSpeedQRevoCurv(RoborockFanPowerCode):
     balanced = 102
     turbo = 103
     max = 104
-    max_plus = 105
+    off = 105
+    custom = 106
+    max_plus = 108
     smart_mode = 110
 
 
@@ -291,6 +295,7 @@ class RoborockMopModeCode(RoborockEnum):
 class RoborockMopModeQRevoCurv(RoborockMopModeCode):
     standard = 300
     deep = 301
+    custom = 302
     deep_plus = 303
     fast = 304
     smart_mode = 306
@@ -373,6 +378,7 @@ class RoborockMopIntensityQRevoCurv(RoborockMopIntensityCode):
     low = 201
     medium = 202
     high = 203
+    custom = 204
     custom_water_flow = 207
     smart_mode = 209
 
@@ -457,6 +463,7 @@ class RoborockDockTypeCode(RoborockEnum):
     s8_maxv_ultra_dock = 10
     qrevo_master_dock = 14
     qrevo_s_dock = 15
+    saros_r10_dock = 16
     qrevo_curv_dock = 17
 
 
