@@ -10,6 +10,7 @@ import time
 
 import aiohttp
 from aiohttp import ContentTypeError, FormData
+from pyrate_limiter import BucketFullException, Duration, Limiter, Rate
 
 from roborock.containers import HomeData, HomeDataRoom, HomeDataScene, ProductResponse, RRiot, UserData
 from roborock.exceptions import (
@@ -21,6 +22,7 @@ from roborock.exceptions import (
     RoborockInvalidUserAgreement,
     RoborockMissingParameters,
     RoborockNoUserAgreement,
+    RoborockRateLimit,
     RoborockTooFrequentCodeRequests,
     RoborockTooManyRequest,
     RoborockUrlException,
@@ -30,16 +32,33 @@ _LOGGER = logging.getLogger(__name__)
 
 
 class RoborockApiClient:
-    def __init__(self, username: str, base_url=None) -> None:
+    _LOGIN_RATES = [
+        Rate(1, Duration.SECOND),
+        Rate(3, Duration.MINUTE),
+        Rate(10, Duration.HOUR),
+        Rate(20, Duration.DAY),
+    ]
+    _HOME_DATA_RATES = [
+        Rate(1, Duration.SECOND),
+        Rate(5, Duration.MINUTE),
+        Rate(15, Duration.HOUR),
+        Rate(40, Duration.DAY),
+    ]
+
+    _login_limiter = Limiter(_LOGIN_RATES)
+    _home_data_limiter = Limiter(_HOME_DATA_RATES)
+
+    def __init__(self, username: str, base_url=None, session: aiohttp.ClientSession | None = None) -> None:
         """Sample API Client."""
         self._username = username
         self._default_url = "https://euiot.roborock.com"
         self.base_url = base_url
         self._device_identifier = secrets.token_urlsafe(16)
+        self.session = session
 
     async def _get_base_url(self) -> str:
         if not self.base_url:
-            url_request = PreparedRequest(self._default_url)
+            url_request = PreparedRequest(self._default_url, self.session)
             response = await url_request.request(
                 "post",
                 "/api/v1/getUrlByEmail",
@@ -113,7 +132,7 @@ class RoborockApiClient:
         ):
             raise RoborockException("Your userdata is missing critical attributes.")
         base_url = user_data.rriot.r.a
-        prepare_request = PreparedRequest(base_url)
+        prepare_request = PreparedRequest(base_url, self.session)
         hid = await self._get_home_id(user_data)
 
         data = FormData()
@@ -151,7 +170,7 @@ class RoborockApiClient:
         ):
             raise RoborockException("Your userdata is missing critical attributes.")
         base_url = user_data.rriot.r.a
-        add_device_request = PreparedRequest(base_url)
+        add_device_request = PreparedRequest(base_url, self.session)
 
         add_device_response = await add_device_request.request(
             "GET",
@@ -174,9 +193,14 @@ class RoborockApiClient:
         return add_device_response["result"]
 
     async def request_code(self) -> None:
+        try:
+            self._login_limiter.try_acquire("login")
+        except BucketFullException as ex:
+            _LOGGER.info(ex.meta_info)
+            raise RoborockRateLimit("Reached maximum requests for login. Please try again later.") from ex
         base_url = await self._get_base_url()
         header_clientid = self._get_header_client_id()
-        code_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
+        code_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
 
         code_response = await code_request.request(
             "post",
@@ -198,10 +222,15 @@ class RoborockApiClient:
                 raise RoborockException(f"{code_response.get('msg')} - response code: {code_response.get('code')}")
 
     async def pass_login(self, password: str) -> UserData:
+        try:
+            self._login_limiter.try_acquire("login")
+        except BucketFullException as ex:
+            _LOGGER.info(ex.meta_info)
+            raise RoborockRateLimit("Reached maximum requests for login. Please try again later.") from ex
         base_url = await self._get_base_url()
         header_clientid = self._get_header_client_id()
 
-        login_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
+        login_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
         login_response = await login_request.request(
             "post",
             "/api/v1/login",
@@ -239,7 +268,7 @@ class RoborockApiClient:
         base_url = await self._get_base_url()
         header_clientid = self._get_header_client_id()
 
-        login_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
+        login_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
         login_response = await login_request.request(
             "post",
             "/api/v1/loginWithCode",
@@ -270,7 +299,7 @@ class RoborockApiClient:
     async def _get_home_id(self, user_data: UserData):
         base_url = await self._get_base_url()
         header_clientid = self._get_header_client_id()
-        home_id_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
+        home_id_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
         home_id_response = await home_id_request.request(
             "get",
             "/api/v1/getHomeDetail",
@@ -285,9 +314,14 @@ class RoborockApiClient:
                 )
             raise RoborockException(f"{home_id_response.get('msg')} - response code: {home_id_response.get('code')}")
 
-        return home_id_response["data"].get("rrHomeId")
+        return home_id_response["data"]["rrHomeId"]
 
     async def get_home_data(self, user_data: UserData) -> HomeData:
+        try:
+            self._home_data_limiter.try_acquire("home_data")
+        except BucketFullException as ex:
+            _LOGGER.info(ex.meta_info)
+            raise RoborockRateLimit("Reached maximum requests for home data. Please try again later.") from ex
         rriot = user_data.rriot
         if rriot is None:
             raise RoborockException("rriot is none")
@@ -296,6 +330,7 @@ class RoborockApiClient:
             raise RoborockException("Missing field 'a' in rriot reference")
         home_request = PreparedRequest(
             rriot.r.a,
+            self.session,
             {
                 "Authorization": self._get_hawk_authentication(rriot, f"/user/homes/{str(home_id)}"),
             },
@@ -311,6 +346,11 @@ class RoborockApiClient:
 
     async def get_home_data_v2(self, user_data: UserData) -> HomeData:
         """This is the same as get_home_data, but uses a different endpoint and includes non-robotic vacuums."""
+        try:
+            self._home_data_limiter.try_acquire("home_data")
+        except BucketFullException as ex:
+            _LOGGER.info(ex.meta_info)
+            raise RoborockRateLimit("Reached maximum requests for home data. Please try again later.") from ex
         rriot = user_data.rriot
         if rriot is None:
             raise RoborockException("rriot is none")
@@ -319,6 +359,7 @@ class RoborockApiClient:
             raise RoborockException("Missing field 'a' in rriot reference")
         home_request = PreparedRequest(
             rriot.r.a,
+            self.session,
             {
                 "Authorization": self._get_hawk_authentication(rriot, "/v2/user/homes/" + str(home_id)),
             },
@@ -332,6 +373,31 @@ class RoborockApiClient:
         else:
             raise RoborockException("home_response result was an unexpected type")
 
+    async def get_home_data_v3(self, user_data: UserData) -> HomeData:
+        """This is the same as get_home_data, but uses a different endpoint and includes non-robotic vacuums."""
+        try:
+            self._home_data_limiter.try_acquire("home_data")
+        except BucketFullException as ex:
+            _LOGGER.info(ex.meta_info)
+            raise RoborockRateLimit("Reached maximum requests for home data. Please try again later.") from ex
+        rriot = user_data.rriot
+        home_id = await self._get_home_id(user_data)
+        if rriot.r.a is None:
+            raise RoborockException("Missing field 'a' in rriot reference")
+        home_request = PreparedRequest(
+            rriot.r.a,
+            {
+                "Authorization": self._get_hawk_authentication(rriot, "/v3/user/homes/" + home_id),
+            },
+        )
+        home_response = await home_request.request("get", "/v3/user/homes/" + home_id)
+        if not home_response.get("success"):
+            raise RoborockException(home_response)
+        home_data = home_response.get("result")
+        if isinstance(home_data, dict):
+            return HomeData.from_dict(home_data)
+        raise RoborockException(f"home_response result was an unexpected type: {home_data}")
+
     async def get_rooms(self, user_data: UserData, home_id: int | None = None) -> list[HomeDataRoom]:
         rriot = user_data.rriot
         if rriot is None:
@@ -342,6 +408,7 @@ class RoborockApiClient:
             raise RoborockException("Missing field 'a' in rriot reference")
         room_request = PreparedRequest(
             rriot.r.a,
+            self.session,
             {
                 "Authorization": self._get_hawk_authentication(rriot, "/v2/user/homes/" + str(home_id)),
             },
@@ -366,6 +433,7 @@ class RoborockApiClient:
             raise RoborockException("Missing field 'a' in rriot reference")
         scenes_request = PreparedRequest(
             rriot.r.a,
+            self.session,
             {
                 "Authorization": self._get_hawk_authentication(rriot, f"/user/scene/device/{str(device_id)}"),
             },
@@ -387,6 +455,7 @@ class RoborockApiClient:
             raise RoborockException("Missing field 'a' in rriot reference")
         execute_scene_request = PreparedRequest(
             rriot.r.a,
+            self.session,
             {
                 "Authorization": self._get_hawk_authentication(rriot, f"/user/scene/{str(scene_id)}/execute"),
             },
@@ -399,7 +468,7 @@ class RoborockApiClient:
         """Gets all products and their schemas, good for determining status codes and model numbers."""
         base_url = await self._get_base_url()
         header_clientid = self._get_header_client_id()
-        product_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
+        product_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
         product_response = await product_request.request(
             "get",
             "/api/v4/product",
@@ -417,7 +486,7 @@ class RoborockApiClient:
     async def download_code(self, user_data: UserData, product_id: int):
         base_url = await self._get_base_url()
         header_clientid = self._get_header_client_id()
-        product_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
+        product_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
         request = {"apilevel": 99999, "productids": [product_id], "type": 2}
         response = await product_request.request(
             "post",
@@ -430,7 +499,7 @@ class RoborockApiClient:
     async def download_category_code(self, user_data: UserData):
         base_url = await self._get_base_url()
         header_clientid = self._get_header_client_id()
-        product_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
+        product_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
         response = await product_request.request(
             "get",
             "api/v1/plugins?apiLevel=99999&type=2",
@@ -442,25 +511,32 @@ class RoborockApiClient:
 
 
 class PreparedRequest:
-    def __init__(self, base_url: str, base_headers: dict | None = None) -> None:
+    def __init__(
+        self, base_url: str, session: aiohttp.ClientSession | None = None, base_headers: dict | None = None
+    ) -> None:
         self.base_url = base_url
         self.base_headers = base_headers or {}
+        self.session = session
 
     async def request(self, method: str, url: str, params=None, data=None, headers=None, json=None) -> dict:
         _url = "/".join(s.strip("/") for s in [self.base_url, url])
         _headers = {**self.base_headers, **(headers or {})}
-        async with aiohttp.ClientSession() as session:
+        close_session = self.session is None
+        session = self.session if self.session is not None else aiohttp.ClientSession()
+        try:
+            async with session.request(method, _url, params=params, data=data, headers=_headers, json=json) as resp:
+                return await resp.json()
+        except ContentTypeError as err:
+            """If we get an error, lets log everything for debugging."""
             try:
-                async with session.request(method, _url, params=params, data=data, headers=_headers, json=json) as resp:
-                    return await resp.json()
-            except ContentTypeError as err:
-                """If we get an error, lets log everything for debugging."""
-                try:
-                    resp_json = await resp.json(content_type=None)
-                    _LOGGER.info("Resp: %s", resp_json)
-                except ContentTypeError as err_2:
-                    _LOGGER.info(err_2)
-                resp_raw = await resp.read()
-                _LOGGER.info("Resp raw: %s", resp_raw)
-                # Still raise the err so that it's clear it failed.
-                raise err
+                resp_json = await resp.json(content_type=None)
+                _LOGGER.info("Resp: %s", resp_json)
+            except ContentTypeError as err_2:
+                _LOGGER.info(err_2)
+            resp_raw = await resp.read()
+            _LOGGER.info("Resp raw: %s", resp_raw)
+            # Still raise the err so that it's clear it failed.
+            raise err
+        finally:
+            if close_session:
+                await session.close()
